# Autogenerated by Astropy's setup.py on 2018-08-02 18:39:18 UTC
from __future__ import unicode_literals
import datetime

version = "2.0.8"
githash = "ca01b3bb62d07d938a526dbee6ff875d9917a9c9"


major = 2
minor = 0
bugfix = 8

release = True
timestamp = datetime.datetime(2018, 8, 2, 18, 39, 18)
debug = False

astropy_helpers_version = "2.0.7"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
